/*(PD)
-----------------------------------------------------------------
	zstest.c - ZeeSpace test
-----------------------------------------------------------------
 * David Olofson, 2005
 *
 * This code is in the Public Domain. No warranty!
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "zeespace.h"

#define	SPEED256	32.0f
#define	SPEED90		15.0f

static void terrain(ZS_Surface *map, float base, float height, float detail,
		int tilesize)
{
	ZS_Pixel px;
	px.r = 160;
	px.g = 128;
	px.b = 96;
	px.a = 255;
	px.i = 256;
	px.z = 0;
	zs_RawFill(ZS_CH_ALL, map, NULL, &px);
	zs_TerrainZ(map, base, height, detail, tilesize);
}


static void drawstuff(ZS_Pipe *p, ZS_Surface *map, int objects)
{
	ZS_Pixel px;
	ZS_State *st = map->state;
	int i;
	objects /= 5;
	px.a = 255;
	px.z = 0;
	for(i = 0; i < objects; ++i)
	{
		int w = 10 + zs_RndGet(st, 80);
		int h = 10 + zs_RndGet(st, 80);
		int x = zs_RndGet(st, map->w - w);
		int y = zs_RndGet(st, map->h - h);
		px.z = 30 * 256 + zs_RndGet(st, 100 * 256);
		px.r = 128 + 20 * sin(i);
		px.g = 128 + 20 * sin(i + 6.25 / 3);
		px.b = 128 + 20 * sin(i + 2 * 6.25 / 3);
		zs_Block(p, map, x, y, w, h, &px);
	}
	for(i = 0; i < objects; ++i)
	{
		int r = 10 + zs_RndGet(st, 40);
		px.z = 50 * 256 + zs_RndGet(st, 100 * 256);
		px.r = 128 + 20 * sin(i);
		px.g = 128 + 20 * sin(i + 6.25 / 3);
		px.b = 128 + 20 * sin(i + 2 * 6.25 / 3);
		zs_Cylinder(p, map, r + zs_RndGet(st, map->w - 2 * r),
				r + zs_RndGet(st, map->h - 2 * r),
				r, &px);
	}
	for(i = 0; i < objects; ++i)
	{
		int r = 10 + zs_RndGet(st, 60);
		int h = 30 * 256 + zs_RndGet(st, 100 * 256);
		px.z = 256 + zs_RndGet(st, 50 * 256);
		px.r = 128 + 20 * sin(i);
		px.g = 128 + 20 * sin(i + 6.25 / 3);
		px.b = 128 + 20 * sin(i + 2 * 6.25 / 3);
		zs_Cone(p, map, r + zs_RndGet(st, map->w - 2 * r),
				r + zs_RndGet(st, map->h - 2 * r),
				r, h, &px);
	}
	for(i = 0; i < objects; ++i)
	{
		int r = 10 + zs_RndGet(st, 60);
		px.r = 128 + 20 * sin(i);
		px.g = 128 + 20 * sin(i + 6.25 / 3);
		px.b = 128 + 20 * sin(i + 2 * 6.25 / 3);
		px.z = 20 * 256 + zs_RndGet(st, 100 * 256);
		zs_Dome(p, map, r + zs_RndGet(st, map->w - 2 * r),
				r + zs_RndGet(st, map->h - 2 * r),
				r, r, &px);
	}
	for(i = 0; i < objects; ++i)
	{
		int r = 10 + zs_RndGet(st, 40);
		int h = 20 + zs_RndGet(st, 100);
		px.z = 256 + zs_RndGet(st, 50 * 256);
		px.r = 128 + 20 * sin(i);
		px.g = 128 + 20 * sin(i + 6.25 / 3);
		px.b = 128 + 20 * sin(i + 2 * 6.25 / 3);
		zs_Beehive(p, map, r + zs_RndGet(st, map->w - 2 * r),
				r + zs_RndGet(st, map->h - 2 * r),
				r, h, &px);
	}
}


static void water(ZS_Surface *map, int sealevel)
{
	ZS_Pixel px;
	px.z = sealevel * 256;
	zs_WaterZ(map, NULL, &px, 96);
}


static void water2(ZS_Surface *map, float altitude, int sealevel)
{
	float afx = sin(altitude * 2 * M_PI / 360);
	ZS_Pixel px;
	px.r = 80;
	px.g = 112;
	px.b = 160;
	px.a = 80;
	px.i = 100 * afx + 100 * afx * afx * afx;
	px.z = sealevel * 256;
	zs_Fog(map, NULL, &px, 48 - 24 * afx);
}


static void fog(ZS_Surface *map, float altitude, float density)
{
	float afx = sin(altitude * 2 * M_PI / 360);
	ZS_Pixel px;
	density = 256 - density;
	px.r = 255;
	px.g = 192;
	px.b = 255;
	px.a = 0;
	px.i = 120 * afx;
	px.z = 150 * 256;
	zs_Fog(map, NULL, &px, density * density / 128 + 100 - 90 * afx);
}


typedef struct
{
	/* System */
	double		t;		/* Current time (s) */
	double		mt;		/* Current morph time (s) */
	int		intro;		/* Intro sequence running */

	/* Display */
	SDL_Surface	*screen;
	int		flags;		/* SDL screen flags */
	int		screenw;
	int		screenh;
	int		tilesize;

	/* Rendering */
	ZS_Surface	*map;		/* "Static" map surface */
	ZS_Surface	*work;		/* Intermediate work surface */
	ZS_Pipe		*drawpipe;	/* Pipeline for drawing shapes */

	/* Flags & selectors */
	int		running;
	int		stuff;
	int		morph;
	int		shadow;
	int		bump;
	int		rebuild;
	int		relight;
	int		graph;
	int		scale;

	/* Angles */
	float		pitch;
	float		altitude;

	/* Ambient light */
	float		ambient;

	/* Levels and other parameters */
	float		sealevel;
	float		fog;
	float		base;
	float		height;
	float		detail;
} STATE;


static void state_Init(STATE *s)
{
	s->t = s->mt = 0;
	s->intro = 0;

	s->flags = SDL_SWSURFACE;
	s->screenw = 640;
	s->screenh = 480;
	s->tilesize = 128;

	s->running = 1;
	s->stuff = 1;
	s->morph = 0;
	s->shadow = 1;
	s->bump = 1;
	s->rebuild = 1;
	s->relight = 1;
	s->graph = 0;
	s->scale = 64;

	s->pitch = 0;
	s->altitude = 0;

	s->ambient = 0.3;

	s->sealevel = 0;
	s->fog = 140;
	s->base = 60;
	s->height = 220;
	s->detail = 0.42;
}


static void start_intro(STATE *s)
{
	s->mt = .8;
	s->pitch = 15;
	s->altitude = 0;
	s->relight = 1;
	s->intro = 1;
}


static void leave_intro(STATE *s)
{
	if(!s->intro)
		return;
	s->intro = 0;
	s->pitch = 90;
	s->altitude = 45;
	s->relight = 1;
}


static void run_intro(STATE *s, float dt)
{
	if(!s->intro)
		return;
	s->pitch += SPEED90 * .7 * dt;
	if(s->pitch > 90)
		s->pitch = 90;
	s->altitude += SPEED90 / 16 * dt;
	s->altitude += (45 - s->altitude) * .1 * dt;;
	s->relight = 1;
	if(s->altitude > 44.9)
		leave_intro(s);
}


static void parse_options(STATE *s, int argc, char *argv[])
{
	int i;
	for(i = 1; i < argc; ++i)
	{
		if(strncmp(argv[i], "-d", 2) == 0)
		{
			s->flags |= SDL_DOUBLEBUF;
			printf("Requesting double buffered display.\n");
		}
		else if(strncmp(argv[i], "-f", 2) == 0)
		{
			s->flags |= SDL_FULLSCREEN;
			printf("Requesting fullscreen display.\n");
		}
		else if(strncmp(argv[i], "-hw", 3) == 0)
		{
			s->flags &= ~SDL_SWSURFACE;
			s->flags |= SDL_HWSURFACE;
			printf("Requesting hardware display surface.\n");
		}
		else if(strncmp(argv[i], "-w", 2) == 0)
		{
			s->screenw = atoi(argv[i] + 2);
			printf("Requested width %d.\n", s->screenw);
		}
		else if(strncmp(argv[i], "-h", 2) == 0)
		{
			s->screenh = atoi(argv[i] + 2);
			printf("Requested height %d.\n", s->screenh);
		}
		else if(strncmp(argv[i], "-t", 2) == 0)
		{
			s->tilesize = atoi(argv[i] + 2);
			printf("Requested tile size %d.\n", s->tilesize);
		}
	}
}


static void check_input(STATE *s, float dt)
{
	SDL_Event event;
	Uint8 *ks = SDL_GetKeyState(NULL);

	/* Action/event keys */
	while(SDL_PollEvent(&event))
	{
		switch(event.type)
		{
		  case SDL_MOUSEBUTTONDOWN:
			s->graph = (s->graph + 1) % 4;
			break;
		  case SDL_KEYDOWN:
			leave_intro(s);
			switch(event.key.keysym.sym)
			{
			  case SDLK_ESCAPE:
				s->running = 0;
				break;
			  case SDLK_F1:
				s->graph = 0;
				break;
			  case SDLK_F2:
				s->graph = 1;
				break;
			  case SDLK_F3:
				s->graph = 2;
				break;
			  case SDLK_F4:
				s->graph = 3;
				break;
			  case SDLK_KP_MULTIPLY:
			  case SDLK_ASTERISK:
				if(s->scale < 256)
					s->scale *= 2;
				break;
			  case SDLK_KP_DIVIDE:
			  case SDLK_SLASH:
				if(s->scale > 2)
					s->scale /= 2;
				break;
			  case SDLK_b:
				s->bump = !s->bump;
				s->relight = 1;
				break;
			  case SDLK_s:
				s->shadow = !s->shadow;
				s->relight = 1;
				break;
			  case SDLK_t:
				s->stuff = !s->stuff;
				s->rebuild = 1;
				break;
			  case SDLK_SPACE:
				s->morph = !s->morph;
				s->rebuild = 1;
				break;
			  default:
				break;
			}
			break;
		}
	}

	/* Control/state keys */
	if(ks[SDLK_PLUS] || ks[SDLK_KP_PLUS])
	{
		s->detail += 0.25 * dt;
		if(s->detail > 2.0)
			s->detail = 2.0;
		s->rebuild = 1;
	}
	else if(ks[SDLK_MINUS] || ks[SDLK_KP_MINUS])
	{
		s->detail -= 0.25 * dt;
		if(s->detail < 0.0)
			s->detail = 0.0;
		s->rebuild = 1;
	}

	if(ks[SDLK_PAGEDOWN])
	{
		s->height -= SPEED256 * dt;
		if(s->height < 0)
			s->height = 0;
		s->rebuild = 1;
	}
	else if(ks[SDLK_PAGEUP])
	{
		s->height += SPEED256 * dt;
		if(s->height > 256)
			s->height = 256;
		s->rebuild = 1;
	}

	if(ks[SDLK_END])
	{
		s->base -= SPEED256 * dt;
		if(s->base < 0)
			s->base = 0;
		s->rebuild = 1;
	}
	else if(ks[SDLK_HOME])
	{
		s->base += SPEED256 * dt;
		if(s->base > 256)
			s->base = 256;
		s->rebuild = 1;
	}

	if(ks[SDLK_LSHIFT] || ks[SDLK_RSHIFT])
	{
		if(ks[SDLK_DELETE])
		{
			s->fog -= SPEED256 * dt;
			if(s->fog < 0)
				s->fog = 0;
			s->rebuild = 1;
		}
		else if(ks[SDLK_INSERT])
		{
			s->fog += SPEED256 * dt;
			if(s->fog > 256)
				s->fog = 256;
			s->rebuild = 1;
		}
	}
	else
	{
		if(ks[SDLK_DELETE])
		{
			s->sealevel -= SPEED256 * dt;
			if(s->sealevel < 0)
				s->sealevel = 0;
			s->rebuild = 1;
		}
		else if(ks[SDLK_INSERT])
		{
			s->sealevel += SPEED256 * dt;
			if(s->sealevel > 255)
				s->sealevel = 255;
			s->rebuild = 1;
		}
	}
	
	if(ks[SDLK_UP])
	{
		s->pitch += SPEED90 * dt;
		if(s->pitch > 90)
			s->pitch = 90;
	}
	else if(ks[SDLK_DOWN])
	{
		s->pitch -= SPEED90 * dt;
		if(s->pitch < 0)
			s->pitch = 0;
	}

	if(ks[SDLK_LEFT])
	{
		s->altitude -= SPEED90 * dt;
		if(s->altitude < 0)
			s->altitude = 0;
		s->relight = 1;
	}
	else if(ks[SDLK_RIGHT])
	{
		s->altitude += SPEED90 * dt;
		if(s->altitude > 90)
			s->altitude = 90;
		s->relight = 1;
	}
}


int main(int argc, char *argv[])
{
	ZS_State *zss;
	STATE s;
	int t1, t2;
	int fps_start, fps_count;
	state_Init(&s);
	SDL_Init(SDL_INIT_VIDEO | SDL_INIT_NOPARACHUTE);
	atexit(SDL_Quit);
	parse_options(&s, argc, argv);

	s.screen = SDL_SetVideoMode(s.screenw, s.screenh, 24, s.flags);
	if(!s.screen)
	{
		fprintf(stderr, "Failed to open screen!\n");
		exit(1);
	}

	SDL_WM_SetCaption("ZeeSpace Test - (C) David Olofson 2005", "ZeeSpace");

	zss = zs_Open();

	/* Create map (landscape + objects) and work (lighting) surfaces */
	s.map = zs_NewSurface(zss, s.screenw + 1, s.screenh + 1);
	s.work = zs_NewSurface(zss, s.screenw + 1, s.screenh + 1);

	/* Set up*/
	s.drawpipe = zs_NewPipe(zss);
	zs_PipeZ(s.drawpipe, ZS_Z_NORMAL | ZS_ZB_ABOVE);
	zs_PipeColor(s.drawpipe, ZS_C_NORMAL);
	zs_PipeAlpha(s.drawpipe, ZS_A_FIXED, 1.0);
	zs_PipeIntensity(s.drawpipe, ZS_I_FIXED, 1.0);
	zs_PipeWrite(s.drawpipe, ZS_CW_RGB | ZS_AW_OFF | ZS_IW_OFF | ZS_ZW_ON);

	start_intro(&s);
	fps_start = SDL_GetTicks();
	fps_count = 0;
	t1 = t2 = fps_start;
	while(s.running)
	{
		float dt = (float)(t2 - t1) * 0.001;

		check_input(&s, dt);
		run_intro(&s, dt);

		if(s.morph)
			s.mt += dt;

		/* Building and rendering! */
		if(s.morph || s.rebuild)
		{
			ZS_Surface *ws;
			if(s.morph)
				ws = s.work;
			else
				ws = s.map;
			zs_RndReset(zss, .5f + sin(s.mt) * .5f);
			terrain(ws, s.base, s.height, s.detail, s.tilesize);
			if(s.stuff)
				drawstuff(s.drawpipe, ws, 30);
			if(s.sealevel >= 1)
				water(ws, s.sealevel);
			s.rebuild = 0;
			s.relight = 1;
		}
		if(s.relight)
		{
			if(!s.morph)
				zs_RawCopy(ZS_CH_ALL, s.map, NULL, s.work, 0, 0);
			else
			{
				ZS_Pixel p;
				p.i = 256;
				zs_RawFill(ZS_CH_I, s.work, NULL, &p);
			}
			if(s.bump)
				zs_BumpMap(s.work, s.altitude, s.ambient);
			if(s.shadow)
				zs_ShadowCast(s.work, s.altitude, s.ambient);
			if(s.sealevel >= 1)
				water2(s.work, s.altitude, s.sealevel);
			if(s.fog)
				fog(s.work, s.altitude, s.fog);
			s.relight = 0;
		}
		switch(s.graph)
		{
		  case 0:
			if(90 == s.pitch)
				zs_Blit2SDL(s.work, s.screen);
			else
				zs_3D2SDL(s.work, s.screen, s.pitch);
			break;
		  case 1:
			zs_Graph2SDL(s.work, s.screen, s.scale);
			break;
		  case 2:
			zs_Z2SDL(s.work, s.screen, s.scale);
			break;
		  case 3:
			zs_I2SDL(s.work, s.screen);
			break;
		}

		/* Display and timing */
		SDL_Flip(s.screen);
		s.t += dt;
		t1 = t2;
		t2 = SDL_GetTicks();
		++fps_count;
	}

	printf("Average fps: %f\n",
			fps_count * 1000.0 / (SDL_GetTicks() - fps_start));

	zs_FreeSurface(s.work);
	zs_FreeSurface(s.map);
	zs_FreePipe(s.drawpipe);
	zs_Close(zss);
	exit(0);
}
